/*@ Message Digests and Authentication Codes.
 *
 * Copyright (c) 2021 - 2023 Steffen Nurpmeso <steffen@sdaoden.eu>.
 * SPDX-License-Identifier: ISC
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */
#ifndef su_MD_H
#define su_MD_H
#include <su/code.h>
su_USECASE_MX_DISABLED
#if defined su_HAVE_MD || defined DOXYGEN /*XXX DOXYGEN bug; ifdef su_HAVE_MD*/
#define su_HEADER
#include <su/code-in.h>
C_DECL_BEGIN
struct su_md;
/* su_md {{{ */
enum su_md_algo{
	su_MD_ALGO_SIPHASH,
	su_MD_ALGO_EXTRA
};
enum su_md_prop{
	su_MD_PROP_ALGO,
	su_MD_PROP_NAME,
	su_MD_PROP_DISPLAY_NAME,
	su_MD_PROP_KEY_SIZE_MIN,
	su_MD_PROP_KEY_SIZE_MAX,
	su_MD_PROP_DIGEST_SIZE_MIN,
	su_MD_PROP_DIGEST_SIZE_MAX,
	su_MD_PROP_BLOCK_SIZE
};
struct su_md_vtbl{
	su_new_fun mdvtbl_new;
	su_del_fun mdvtbl_del;
	up (*mdvtbl_property)(void const *self, enum su_md_prop prop);
	s32 (*mdvtbl_setup)(void *self, void const *key, uz key_len, uz digest_size);
	void (*mdvtbl_update)(void *self, void const *dat, uz dat_len);
	void (*mdvtbl_end)(void *self, void *store);
};
struct su_md{
	struct su_md_vtbl const *md_vtbl;
	void *md_vp;
};
EXPORT s32 su__md_install(char const *name, struct su_md_vtbl const *vtblp, su_new_fun cxx_it, u32 estate);
EXPORT boole su__md_uninstall(char const *name, struct su_md_vtbl const *vtblp, su_new_fun cxx_it);
EXPORT struct su_md *su_md_new_by_algo(enum su_md_algo algo, u32 estate);
EXPORT struct su_md *su_md_new_by_name(char const *name, u32 estate);
EXPORT void su_md_del(struct su_md *self);
INLINE up su_md_property(struct su_md const *self, enum su_md_prop prop){
	ASSERT(self);
	return (*self->md_vtbl->mdvtbl_property)(self->md_vp, prop);
}
INLINE char const *su_md_name(struct su_md const *self){
	ASSERT(self);
	return R(char const*,su_md_property(self, su_MD_PROP_NAME));
}
INLINE char const *su_md_display_name(struct su_md const *self){
	ASSERT(self);
	return R(char const*,su_md_property(self, su_MD_PROP_DISPLAY_NAME));
}
INLINE s32 su_md_setup(struct su_md *self, void const *key, uz key_len, uz digest_size){
	ASSERT(self);
	ASSERT_RET(key_len == 0 || key != NIL, -su_ERR_FAULT);
	return (*self->md_vtbl->mdvtbl_setup)(self->md_vp, key, key_len, digest_size);
}
INLINE void su_md_update(struct su_md *self, void const *dat, uz dat_len){
	ASSERT(self);
	ASSERT_RET_VOID(dat_len == 0 || dat != NIL);
	if(dat_len > 0)
		(*self->md_vtbl->mdvtbl_update)(self->md_vp, dat, dat_len);
}
INLINE void su_md_end(struct su_md *self, void *store){
	ASSERT(self);
	ASSERT_RET_VOID(store != NIL);
	(*self->md_vtbl->mdvtbl_end)(self->md_vp, store);
}
INLINE s32 su_md_install(char const *name, struct su_md_vtbl const *vtblp, u32 estate){
	ASSERT_RET(name != NIL, -su_ERR_FAULT);
	ASSERT_RET(vtblp != NIL, -su_ERR_FAULT);
	return su__md_install(name, vtblp, NIL, estate);
}
INLINE boole su_md_uninstall(char const *name, struct su_md_vtbl const *vtblp){
	ASSERT_RET(name != NIL, FAL0);
	ASSERT_RET(vtblp != NIL, FAL0);
	return su__md_uninstall(name, vtblp, NIL);
}
/* }}} */
C_DECL_END
#include <su/code-ou.h>
#endif /* su_HAVE_MD */
#endif /* su_MD_H */
/* s-itt-mode */
