/*@ Mutual exclusion device.
 *
 * Copyright (c) 2001 - 2023 Steffen Nurpmeso <steffen@sdaoden.eu>.
 * SPDX-License-Identifier: ISC
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */
#ifndef su_MUTEX_H
#define su_MUTEX_H
#include <su/code.h>
su_USECASE_MX_DISABLED
#include <su/atomic.h>
#include <su/thread.h>
#include <su/x-mutex.h> /* 1. */
#define su_HEADER
#include <su/code-in.h>
C_DECL_BEGIN
struct su_mutex;
/* mutex {{{ */
#if !defined su_HAVE_MT || su_ATOMIC_IS_REAL
# define su__MUTEX_SPIN
#endif
enum su_mutex_flags{
	su_MUTEX_FLAT = 1u<<0,
	su_MUTEX_INIT = 1u<<1
};
#if DVLDBGOR(1, 0)
enum su__mutex_xfn{
	su__MUTEX_GUT,
	su__MUTEX_LOCK,
	su__MUTEX_TRYLOCK,
	su__MUTEX_UNLOCK_NOLOCK,
	su__MUTEX_UNLOCK
};
#endif
struct su_mutex{
	struct{
		MT( u8 os[su__MUTEX_IMPL_SIZE]; )
#ifndef su__MUTEX_SPIN
		MT( u8 os_lck[su__MUTEX_IMPL_SIZE]; )
#endif
		u8 ATOMIC lck; /* Iff su__MUTEX_SPIN */
		u8 flags;
		u16 line; /* dbg */
		u32 count;
		struct su_thread *owner;
		struct su_thread *waiters;
		char const *file; /* dbg */
		char const *name; /* Last field! */
	}
#ifdef su__MUTEX_IMPL_ALIGNMENT
		su_CC_ALIGNED(su__MUTEX_IMPL_ALIGNMENT)
# undef su__MUTEX_IMPL_ALIGNMENT
#endif
		mtx_;
};
#define su__MUTEX_ARGS_DECL su_DVLDBG(su_COMMA char const *file su_COMMA u32 line)
#define su__MUTEX_ARGS su_DVLDBG(su_COMMA file su_COMMA line)
#define su__MUTEX_ARGS_INJ su_DVLDBG(su_COMMA __FILE__ su_COMMA __LINE__)
/* */
#ifdef DOXYGEN
# define su_MUTEX_I9R(NAME_OR_NIL)
# define su_MUTEX_FLAT_I9R(NAME_OR_NIL)
#else
# ifdef su_HAVE_MT
#  ifdef su__MUTEX_SPIN
#   define su__MUTEX_IMPL_I9R() {0,},
#  else
#   define su__MUTEX_IMPL_I9R() {0,}, {0,},
#  endif
# else
#  define su__MUTEX_IMPL_I9R()
# endif
# define su__MUTEX_I9R(FLAGS,NAME_OR_NIL) \
	{{su__MUTEX_IMPL_I9R() \
		su_FIELD_INITN(lck) 0,\
		su_FIELD_INITN(flags) FLAGS,\
		su_FIELD_INITN(line) 0,\
		su_FIELD_INITN(count) 0,\
		su_FIELD_INITN(owner) su_NIL,\
		su_FIELD_INITN(waiters) su_NIL,\
		su_FIELD_INITN(file) su_NIL,\
		su_FIELD_INITN(name) NAME_OR_NIL}}
# define su_MUTEX_I9R(NAME_OR_NIL) su__MUTEX_I9R(0, NAME_OR_NIL)
# define su_MUTEX_FLAT_I9R(NAME_OR_NIL) su__MUTEX_I9R(su_MUTEX_FLAT, NAME_OR_NIL)
#endif /* !DOXYGEN */
#define su_MUTEX_LOCK(SELF) su_DVLDBGOR(su_mutex_lock_dbg,su_mutex_lock)(SELF su__MUTEX_ARGS_INJ)
#define su_MUTEX_LOCK_TS(SELF,TSP) su_DVLDBGOR(su_mutex_lock_ts_dbg,su_mutex_lock_ts)(SELF, TSP su__MUTEX_ARGS_INJ)
#define su_MUTEX_TRYLOCK(SELF) su_DVLDBGOR(su_mutex_trylock_dbg,su_mutex_trylock)(SELF su__MUTEX_ARGS_INJ)
#define su_MUTEX_TRYLOCK_TS(SELF,TSP) su_DVLDBGOR(su_mutex_trylock_ts_dbg,su_mutex_trylock_ts)(SELF, TSP su__MUTEX_ARGS_INJ)
#define su_MUTEX_UNLOCK(SELF) su_DVLDBGOR(su_mutex_unlock_dbg,su_mutex_unlock)(SELF su__MUTEX_ARGS_INJ)
/* */
#if DVLDBGOR(1, 0)
EXPORT boole su__mutex_check(struct su_mutex *self, enum su__mutex_xfn mf, struct su_thread *tsp, char const *file, u32 line);
#endif
#include <su/x-mutex.h> /* 2. */
#ifndef su__MUTEX_HAVE_FUNS
# ifndef su_HAVE_MT
#  error Combination will not work
# endif
EXPORT s32 su__mutex_os_create(struct su_mutex *self, u32 estate);
EXPORT void su__mutex_os_gut(struct su_mutex *self);
EXPORT void su__mutex_lock(struct su_mutex *self, struct su_thread *tsp  su__MUTEX_ARGS_DECL);
EXPORT boole su__mutex_trylock(struct su_mutex *self, struct su_thread *tsp  su__MUTEX_ARGS_DECL);
EXPORT void su__mutex_unlock(struct su_mutex *self  su__MUTEX_ARGS_DECL);
#endif
INLINE s32 su_mutex_create(struct su_mutex *self, char const *name_or_nil, u32 estate){
	s32 rv;
	ASSERT(self);
	UNUSED(estate);
	FIELD_RANGE_ZERO(struct su_mutex, self, mtx_.lck, mtx_.file);
	self->mtx_.name = name_or_nil;
	rv = su_STATE_NONE;
	MT(
		if((rv = su__mutex_os_create(self, estate)) == su_STATE_NONE)
			self->mtx_.flags |= su_MUTEX_INIT;
	)
	return rv;
}
INLINE void su_mutex_gut(struct su_mutex *self){
	ASSERT(self);
	UNUSED(self);
	DVLDBG( if(su__mutex_check(self, su__MUTEX_GUT, NIL su__MUTEX_ARGS_INJ)) )
		MT( if(self->mtx_.flags & su_MUTEX_INIT) su__mutex_os_gut(self) )
			do{}while(0);
}
INLINE void su_mutex_disable_recursion(struct su_mutex *self){
	ASSERT(self);
	UNUSED(self);
	ASSERT(self->mtx_.owner == NIL);
	self->mtx_.flags |= su_MUTEX_FLAT;
}
INLINE char const *su_mutex_name(struct su_mutex const *self){
	ASSERT(self);
	return self->mtx_.name;
}
INLINE char const *su_mutex_file(struct su_mutex const *self){
	ASSERT(self);
	UNUSED(self);
	return DVLDBGOR(self->mtx_.file, NIL);
}
INLINE u32 su_mutex_line(struct su_mutex const *self){
	ASSERT(self);
	UNUSED(self);
	return DVLDBGOR(self->mtx_.line, 0);
}
INLINE boole su_mutex_is_locked(struct su_mutex const *self){
	ASSERT(self);
	return (self->mtx_.owner != NIL);
}
INLINE void su_mutex_lock_ts_dbg(struct su_mutex *self, struct su_thread *tsp, char const *file, u32 line){
	ASSERT(self);
	ASSERT_EXEC(tsp == su_thread_self(), tsp = su_thread_self());
	UNUSED(file);
	UNUSED(line);
	su__mutex_lock(self, tsp  su__MUTEX_ARGS);
}
INLINE void su_mutex_lock_ts(struct su_mutex *self, struct su_thread *tsp){
	ASSERT(self);
	ASSERT_EXEC(tsp == su_thread_self(), tsp = su_thread_self());
	su_mutex_lock_ts_dbg(self, tsp, __FILE__, __LINE__);
}
INLINE void su_mutex_lock_dbg(struct su_mutex *self, char const *file, u32 line){
	ASSERT(self);
	su_mutex_lock_ts_dbg(self, su_thread_self(), file, line);
}
INLINE void su_mutex_lock(struct su_mutex *self) {su_mutex_lock_dbg(self, __FILE__, __LINE__);}
INLINE boole su_mutex_trylock_ts_dbg(struct su_mutex *self, struct su_thread *tsp, char const *file, u32 line){
	ASSERT(self);
	ASSERT_EXEC(tsp == su_thread_self(), tsp = su_thread_self());
	UNUSED(file);
	UNUSED(line);
	return su__mutex_trylock(self, tsp su__MUTEX_ARGS);
}
INLINE boole su_mutex_trylock_ts(struct su_mutex *self, struct su_thread *tsp){
	ASSERT(self);
	ASSERT_EXEC(tsp == su_thread_self(), tsp = su_thread_self());
	return su_mutex_trylock_ts_dbg(self, tsp, __FILE__, __LINE__);
}
INLINE boole su_mutex_trylock(struct su_mutex *self){
	ASSERT(self);
	return su_mutex_trylock_ts_dbg(self, su_thread_self(), __FILE__, __LINE__);
}
INLINE boole su_mutex_trylock_dbg(struct su_mutex *self, char const *file, u32 line){
	ASSERT(self);
	return su_mutex_trylock_ts_dbg(self, su_thread_self(), file, line);
}
INLINE void su_mutex_unlock_dbg(struct su_mutex *self, char const *file, u32 line){
	ASSERT(self);
	UNUSED(file);
	UNUSED(line);
	su__mutex_unlock(self su__MUTEX_ARGS);
}
INLINE void su_mutex_unlock(struct su_mutex *self){
	ASSERT(self);
	su_mutex_unlock_dbg(self, __FILE__, __LINE__);
}
/* }}} */
#include <su/x-mutex.h> /* 3. */
#ifndef su_SOURCE_MUTEX
# undef su__MUTEX_SPIN
# undef su__MUTEX_ARGS_DECL
# undef su__MUTEX_ARGS
# undef su__MUTEX_HAVE_FUNS
#endif
C_DECL_END
#include <su/code-ou.h>
#endif /* su_MUTEX_H */
/* s-itt-mode */
