/*@ Spinning mutual exclusion device.
 *
 * Copyright (c) 2001 - 2023 Steffen Nurpmeso <steffen@sdaoden.eu>.
 * SPDX-License-Identifier: ISC
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */
#ifndef su_SPINLOCK_H
#define su_SPINLOCK_H
#include <su/code.h>
su_USECASE_MX_DISABLED
#include <su/atomic.h>
#include <su/thread.h>
#if !defined su_HAVE_MT || su_ATOMIC_IS_REAL
# define su__SPINLOCK_IS
#else
# include <su/mutex.h>
#endif
#include <su/x-spinlock.h> /* 1. */
#define su_HEADER
#include <su/code-in.h>
C_DECL_BEGIN
struct su_spinlock;
/* spinlock {{{ */
#if defined su__SPINLOCK_IS && DVLDBGOR(1, 0)
# define su__SPINLOCK_DBG
#endif
#ifdef su__SPINLOCK_DBG
enum su__spinlock_xfn{
	su__SPIN_DTOR,
	su__SPIN_LOCK,
	su__SPIN_TRYLOCK,
	su__SPIN_UNLOCK
};
#endif
struct su_spinlock{
#ifdef su__SPINLOCK_IS
	up ATOMIC sl_lck; /* With __SPINLOCK_DBG a thread*, otherwise a bool */
	DVLDBG( char const *sl_name; )
#else
	struct su_mutex sl_mtx;
#endif
};
#if defined su__SPINLOCK_IS || defined DOXYGEN
# define su_SPINLOCK_I9R(DBG_NAME_OR_NIL) {0 su_DVLDBG(su_COMMA DBG_NAME_OR_NIL)}
#else
# define su_SPINLOCK_I9R(DBG_NAME_OR_NIL) su_MUTEX_FLAT_I9R(DBG_NAME_OR_NIL)
#endif
#ifdef su__SPINLOCK_DBG
EXPORT boole su__spinlock_check(struct su_spinlock *self, enum su__spinlock_xfn slf, up v);
#endif
#include <su/x-spinlock.h> /* 2. */
INLINE s32 su_spinlock_create(struct su_spinlock *self, char const *dbg_name_or_nil, u32 estate){
	s32 rv;
	ASSERT(self);
	UNUSED(dbg_name_or_nil);
	UNUSED(estate);
#ifdef su__SPINLOCK_IS
	self->sl_lck = 0;
	DVLDBG( self->sl_name = dbg_name_or_nil; )
	rv = su_STATE_NONE;
#else
	rv = su_mutex_create(&self->sl_mtx, dbg_name_or_nil, estate);
#endif
	return rv;
}
INLINE void su_spinlock_gut(struct su_spinlock *self){
	ASSERT(self);
#ifdef su__SPINLOCK_IS
	UNUSED(self);
	DVLDBG( su__spinlock_check(self, su__SPIN_DTOR, 0); )
#else
	su_mutex_gut(&self->sl_mtx);
#endif
}
INLINE void su_spinlock_lock(struct su_spinlock *self){
	ASSERT(self);
#ifdef su__SPINLOCK_IS
	/* C99 */{
		up v = DVLDBGOR(R(up,su_thread_self()), 1);
		DVLDBG( if(!su__spinlock_check(self, su__SPIN_LOCK, v)) return; )
		su__spinlock_lock(self, v);
	}
#else
	su_mutex_lock(&self->sl_mtx);
#endif
}
INLINE boole su_spinlock_trylock(struct su_spinlock *self){
	ASSERT(self);
#ifdef su__SPINLOCK_IS
	/* C99 */{
		up v = DVLDBGOR(R(up,su_thread_self()), 1);
		DVLDBG( if(!su__spinlock_check(self, su__SPIN_TRYLOCK, v)) return FAL0; )
		return su__spinlock_trylock(self, v);
	}
#else
	return su_mutex_trylock(&self->sl_mtx);
#endif
}
INLINE void su_spinlock_unlock(struct su_spinlock *self){
	ASSERT(self);
#ifdef su__SPINLOCK_IS
	DVLDBG( if(!su__spinlock_check(self, su__SPIN_UNLOCK, 0)) return; )
	su__spinlock_unlock(self);
#else
	su_mutex_unlock(&self->sl_mtx);
#endif
}
/* }}} */
#include <su/x-spinlock.h> /* 3. */
#ifndef su_SOURCE_SPINLOCK
# undef su__SPINLOCK_IS
# undef su__SPINLOCK_DBG
#endif
C_DECL_END
#include <su/code-ou.h>
#endif /* su_SPINLOCK_H */
/* s-itt-mode */
